@@ -62,21 +62,26 @@ class geneticalgorithm2():
                 algorithm over iterations
 
     '''
+    
+    default_params = {
+        'max_num_iteration': None,
+        'population_size':100,
+        'mutation_probability':0.1,
+        'elit_ratio': 0.01,
+        'crossover_probability': 0.5,
+        'parents_portion': 0.3,
+        'crossover_type':'uniform',
+        'mutation_type': 'uniform_by_center',
+        'max_iteration_without_improv':None
+        }
+    
+    
     #############################################################
     def __init__(self, function, dimension, variable_type='bool', \
                  variable_boundaries=None,\
                  variable_type_mixed=None, \
                  function_timeout=10,\
-                 algorithm_parameters={'max_num_iteration': None,\
-                                       'population_size':100,\
-                                       'mutation_probability':0.1,\
-                                       'elit_ratio': 0.01,\
-                                       'crossover_probability': 0.5,\
-                                       'parents_portion': 0.3,\
-                                       'crossover_type':'uniform',\
-                                       'max_iteration_without_improv':None}):
-
-
+                 algorithm_parameters = default_params):
         '''
         @param function <Callable> - the given objective function to be minimized
         NOTE: This implementation minimizes the given objective function. 
@@ -116,7 +121,8 @@ class geneticalgorithm2():
             @ elit_ration <float in [0,1]>
             @ crossover_probability <float in [0,1]>
             @ parents_portion <float in [0,1]>
-            @ crossover_type <string> - Default is 'uniform'; 'one_point' or 'two_point' are other options
+            @ crossover_type <string/function> - Default is 'uniform'; 'one_point' or 'two_point' (not only) are other options
+            @ mutation_type <string/function> - Default is 'uniform_by_x'; see GitHub to check other options
             @ max_iteration_without_improv <int> - maximum number of successive iterations without improvement. If None it is ineffective
         
         for more details and examples of implementation please visit:
@@ -166,6 +172,8 @@ class geneticalgorithm2():
                 
 
             self.var_type = variable_type_mixed
+            
+        self.set_crossover_and_mutations(algorithm_parameters['crossover_type'], algorithm_parameters['mutation_type'])
         #############################################################
         # input variables' boundaries 
 
@@ -196,6 +204,8 @@ class geneticalgorithm2():
         # input algorithm's parameters
         
         self.param = algorithm_parameters
+        self.param.update({key:val for key, val in geneticalgorithm2.default_params.items() if key not in list((algorithm_parameters.keys()))})
+        
         
         self.pop_s = int(self.param['population_size'])
         
@@ -243,10 +253,6 @@ class geneticalgorithm2():
         else:
             self.iterate = int(self.param['max_num_iteration'])
         
-        self.c_type = self.param['crossover_type']
-        assert (self.c_type in ['uniform','one_point','two_point']),\
-        "\n crossover_type must 'uniform', 'one_point', or 'two_point' Enter string" 
-        
         
         self.stop_mniwi=False ## what
         if self.param['max_iteration_without_improv'] == None or self.param['max_iteration_without_improv'] < 1:
@@ -255,6 +261,41 @@ class geneticalgorithm2():
             self.mniwi = int(self.param['max_iteration_without_improv'])
 
         
+
+    def set_crossover_and_mutations(self, crossover, mutation):
+        
+        if type(crossover) == str:
+            if crossover == 'one_point':
+                self.crossover = Crossover.one_point()
+            elif crossover == 'two_point':
+                self.crossover = Crossover.two_point()
+            elif crossover == 'uniform':
+                self.crossover = Crossover.uniform()
+            elif crossover == 'segment':
+                self.crossover = Crossover.segment()
+            elif crossover == 'shuffle':
+                self.crossover = Crossover.shuffle()
+            else:
+                raise Exception(f"unknown type of crossover: {crossover}")
+        else:
+            self.crossover = crossover 
+        
+        if type(mutation) == str:
+            if mutation == 'uniform_by_x':
+                self.real_mutation = Mutations.uniform_by_x()
+            elif mutation == 'uniform_by_center':
+                self.real_mutation = Mutations.uniform_by_center()
+            elif mutation == 'gauss_by_center':
+                self.real_mutation = Mutations.gauss_by_center()
+            elif mutation == 'gauss_by_x':
+                self.real_mutation = Mutations.gauss_by_x()
+            else:
+                raise Exception(f"unknown type of mutation: {mutation}")
+        else:
+            self.real_mutation = mutation
+
+
+
         ############################################################# 
     def run(self, no_plot = False, set_function = None, apply_function_to_parents = False, start_generation = {'variables':None, 'scores': None}):
         """
@@ -321,6 +362,9 @@ class geneticalgorithm2():
         #############################################################
         # Report
         self.report = []
+        self.report_min = []
+        self.report_average = []
+        
         self.test_obj = obj
         self.best_variable = var.copy()
         self.best_function = obj
@@ -349,6 +393,8 @@ class geneticalgorithm2():
             # Report
 
             self.report.append(pop[0, self.dim])
+            self.report_min.append(pop[-1, self.dim])
+            self.report_average.append(np.mean(pop[:, self.dim]))
     
             ##############################################################         
             # Normalizing objective function 
@@ -410,7 +456,7 @@ class geneticalgorithm2():
                 pvar1 = ef_par[r1,:self.dim].copy()
                 pvar2 = ef_par[r2,:self.dim].copy()
                 
-                ch1, ch2 = self.cross(pvar1, pvar2, self.c_type)
+                ch1, ch2 = self.crossover(pvar1, pvar2)
                 
                 ch1 = self.mut(ch1)
                 ch2 = self.mutmidle(ch2, pvar1, pvar2)               
@@ -457,6 +503,8 @@ class geneticalgorithm2():
         # Report
 
         self.report.append(pop[0,self.dim])
+        self.report_min.append(pop[-1, self.dim])
+        self.report_average.append(np.mean(pop[:, self.dim]))
         
         
  
@@ -483,7 +531,7 @@ class geneticalgorithm2():
                              ' maximum number of iterations without improvement was met!')
 ##############################################################################         
 
-    def plot_results(self):
+    def plot_results(self, show_mean = False):
         """
         Simple plot of self.report (if not empty)
         """
@@ -491,43 +539,26 @@ class geneticalgorithm2():
             sys.stdout.write("No results to plot!\n")
             return
 
-        re=np.array(self.report)
-        plt.plot(re)
-        plt.xlabel('Iteration')
-        plt.ylabel('Objective function')
+        bests = np.array(self.report)
+        means = np.array(self.report_average)
+        
+        if show_mean: plt.plot(means, color = 'red', label = 'mean by generation', linewidth = 1)
+        plt.plot(bests, color = 'blue', label = 'best of generation')
+        
+        plt.xlabel('Generation')
+        plt.ylabel('Minimized function')
         plt.title('Genetic Algorithm')
+        plt.legend()
         plt.show()
 
-##############################################################################         
-    def cross(self, x, y, c_type):
-         
-        ofs1=x.copy()
-        ofs2=y.copy()
-        
 
-        if c_type=='one_point':
-            ran=np.random.randint(0, self.dim)
-            ofs1[:ran] = y[:ran]
-            ofs2[:ran] = x[:ran]
-  
-        elif c_type=='two_point':
-                
-            ran1=np.random.randint(0, self.dim)
-            ran2=np.random.randint(ran1, self.dim)
-            
-            ofs1[ran1:ran2] = y[ran1:ran2]
-            ofs2[ran1:ran2] = x[ran1:ran2]
-              
-        elif c_type=='uniform':
-            ran = np.random.random(self.dim) < 0.5
-            ofs1[ran] = y[ran]
-            ofs2[ran] = x[ran]
-                   
-        
-        return (ofs1, ofs2)
+
 ###############################################################################  
     
     def mut(self,x):
+        """
+        just mutation
+        """
         
         for i in self.integers[0]:
             if np.random.random() < self.prob_mut:
@@ -536,14 +567,17 @@ class geneticalgorithm2():
         
 
         for i in self.reals[0]:                
-            if np.random.random() < self.prob_mut:   
-               x[i]=np.random.uniform(self.var_bound[i][0], self.var_bound[i][1]) #self.var_bound[i][0]+np.random.random()*\
-                #(self.var_bound[i][1]-self.var_bound[i][0])    
+            if np.random.random() < self.prob_mut:
+                x[i] = self.real_mutation(x[i], self.var_bound[i][0], self.var_bound[i][1]) 
             
         return x
 
 ###############################################################################
     def mutmidle(self, x, p1, p2):
+        """
+        mutation oriented on parents
+        """
+        
         for i in self.integers[0]:
 
             if np.random.random() < self.prob_mut:
@@ -619,6 +653,189 @@ class geneticalgorithm2():
         return func
             
 ###############################################################################
+
+
+
+
+
+class Crossover:
+    
+    def get_copies(x, y):
+        return x.copy(), y.copy()
+    
+    def one_point():
+        
+        def func(x, y):
+            ofs1, ofs2 = Crossover.get_copies(x, y)
+        
+            ran=np.random.randint(0, x.size)
             
+            ofs1[:ran] = y[:ran]
+            ofs2[:ran] = x[:ran]
+            
+            return ofs1, ofs2
+        return func
+    
+    def two_point():
+        
+        def func(x, y):
+            ofs1, ofs2 = Crossover.get_copies(x, y)
+        
+            ran1=np.random.randint(0, x.size)
+            ran2=np.random.randint(ran1, x.size)
+            
+            ofs1[ran1:ran2] = y[ran1:ran2]
+            ofs2[ran1:ran2] = x[ran1:ran2]
+              
+            
+            return ofs1, ofs2
+        return func
+    
+    def uniform():
+        
+        def func(x, y):
+            ofs1, ofs2 = Crossover.get_copies(x, y)
+        
+            ran = np.random.random(x.size) < 0.5
+            ofs1[ran] = y[ran]
+            ofs2[ran] = x[ran]
+              
+            return ofs1, ofs2
+        
+        return func
+    
+    def segment(prob = 0.6):
+        
+        def func(x, y):
+            
+            ofs1, ofs2 = Crossover.get_copies(x, y)
+            
+            p = np.random.random(x.size) < prob
+            
+            for i, val in enumerate(p):
+                if val:
+                    ofs1[i], ofs2[i] = ofs2[i], ofs1[i]
+            
+            return ofs1, ofs2
+        
+        return func
+    
+    def shuffle():
+        
+        def func(x, y):
+            
+            ofs1, ofs2 = Crossover.get_copies(x, y)
+            
+            index = np.random.choice(np.arange(0, x.size), x.size, replace = False)
+            
+            ran = np.random.randint(0, x.size)
+            
+            for i in range(ran):
+                ind = index[i]
+                ofs1[ind] = y[ind]
+                ofs2[ind] = x[ind]
+            
+            return ofs1, ofs2
+            
+        return func
+    
+    #
+    #
+    # ONLY FOR REAL VARIABLES
+    #
+    #
+    
+    def arithmetic():
+        
+        def func(x, y):
+            b = np.random.random()
+            a = 1-b
+            return a*x + b*y, a*y + b*x
+        
+        return func
+    
+    
+    def mixed(alpha = 0.5):
+        
+        def func(x,y):
+            
+            a = np.empty(x.size)
+            b = np.empty(y.size)
+            
+            x_min = np.minimum(a, b)
+            x_max = np.maximum(a, b)
+            delta = alpha*(x_max-x_min)
+            
+            for i in range(x.size):
+                a[i] = np.random.uniform(x_min[i] - delta[i], x_max[i] + delta[i])
+                b[i] = np.random.uniform(x_min[i] - delta[i], x_max[i] + delta[i])
+            
+            return a, b
+        
+        return func
+        
+
+class Mutations:
+    
+    def uniform_by_x():
+        
+        def func(x, left, right):
+            alp = min(x - left, right - x)
+            return np.random.uniform(x - alp, x + alp)
+        return func
+    
+    
+    def uniform_by_center():
+        
+        def func(x, left, right):
+            return np.random.uniform(left, right)
+        
+        return func
+    
+    def gauss_by_x(sd = 0.3):
+        """
+        gauss mutation with x as center and sd*length_of_zone as std
+        """
+        def func(x, left, right):
+            std = sd * (right - left)
+            return max(left, min(right, np.random.normal(loc = x, scale = std)))
+        
+        return func
+    
+    def gauss_by_center(sd = 0.3):
+        """
+        gauss mutation with (left+right)/2 as center and sd*length_of_zone as std
+        """
+        def func(x, left, right):
+            std = sd * (right - left)
+            return max(left, min(right, np.random.normal(loc = (left+right)*0.5, scale = std)))
+        
+        return func
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
             
             
\ No newline at end of file
